/*
 * Mille Bornes
 */

#include <stdio.h>


#ifdef	__STDC__
#define	PROTO(a)	a
#else
#define	PROTO(a)	()
#endif


#define MAGIC		73659171 /* magic number for saved game validation */
#define HANDSIZE	6	/* number of cards in hand */
#define CARDCOUNT	101	/* total number of cards in game */
#define MAX200CARDS	2	/* maximum number of 200 mile cards */
#define PLAYERINDEX	0	/* index of human player */
#define MACHINEINDEX	1	/* index of computer player */


typedef	int	BOOL;

#define TRUE	((BOOL) 1)
#define FALSE	((BOOL) 0)


/*
 * Status bits for game when a hand is over.  Zero status indicates
 * current hand is still in progress.  These bits are or'd together.
 */
typedef	int	STATUS;

#define ST_TRIPDONE	0x01	/* some player reached the mileage goal */
#define ST_HANDOVER	0x02	/* the current hand is finished */
#define ST_HANDWON	0x04	/* the player won the current hand */
#define ST_HANDTIED	0x08	/* the player tied the current hand */
#define ST_GAMEOVER	0x10	/* the current game is finished */
#define ST_GAMEWON	0x20	/* the player won the current game */


/*
 * Scoring values and special mileage values
 */
#define S_SAFETY	100	/* per safety played */
#define S_ALLSAFETY	700	/* all four safeties played */
#define S_COUPFOURRE	300	/* per coup fourre */
#define S_TRIPDONE	400	/* trip completed */
#define S_DELAYED	300	/* delayed action */
#define S_SAFETRIP	300	/* safe trip (no 200's) */
#define S_EXTENSION	200	/* extension */
#define S_SHUTOUT	500	/* shut out */

#define S_LIMIT		50	/* speed limit */
#define S_CAREFUL	100	/* mileage left when become careful */
#define S_MAXMILE	200	/* maximum mileage card */
#define S_SHORTGOAL	700	/* mileage goal for short hand */
#define S_LONGGOAL	1000	/* mileage goal for extended hand */
#define S_GAMEGOAL	5000	/* points needed to win game */


/*
 * Types of cards.
 * The order of these definitions is important.
 */
typedef	int	CARD;

#define TC_NONE		0	/* not really a card (MUST BE ZERO) */
#define TC_25MILE	1	/* mileage cards */
#define TC_50MILE	2
#define TC_75MILE	3
#define TC_100MILE	4
#define TC_200MILE	5
#define TC_OUTOFGAS	6	/* hazard cards */
#define TC_FLATTIRE	7
#define TC_ACCIDENT	8
#define TC_STOP		9
#define TC_SPEEDLIMIT	10
#define TC_GASOLINE	11	/* remedy cards */
#define TC_SPARETIRE	12
#define TC_REPAIRS	13
#define TC_ROLL		14
#define TC_ENDOFLIMIT	15
#define TC_EXTRATANK	16	/* safety cards */
#define TC_PUNCTUREPROOF 17
#define TC_DRIVINGACE	18
#define TC_RIGHTOFWAY	19

#define CARDTYPES	20	/* total number of card types */
#define MILETYPES	5	/* number of mileage types */
#define SAFETYPES	4	/* number of safety card types */


#define isbetween(n,minval,maxval) (((n) >= (minval)) && ((n) <= (maxval)))
#define ismileage(n)	isbetween(n, TC_25MILE, TC_200MILE)
#define ishazard(n)	isbetween(n, TC_OUTOFGAS, TC_SPEEDLIMIT)
#define isremedy(n)	isbetween(n, TC_GASOLINE, TC_ENDOFLIMIT)
#define issafety(n)	isbetween(n, TC_EXTRATANK, TC_RIGHTOFWAY)
#define iscard(n)	((n) != TC_NONE)
#define isnocard(n)	((n) == TC_NONE)

#define mileageindex(n)	((n) - TC_25MILE)
#define hazardindex(n)	((n) - TC_OUTOFGAS)
#define remedyindex(n)	((n) - TC_GASOLINE)
#define safetyindex(n)	((n) - TC_EXTRATANK)

#define hazardtoremedy(n)	((n) + (TC_GASOLINE - TC_OUTOFGAS))
#define remedytohazard(n)	((n) + (TC_OUTOFGAS - TC_GASOLINE))
#define hazardtosafety(n)	((((n)<=TC_STOP)?n:TC_STOP)+(TC_EXTRATANK-TC_OUTOFGAS))
#define remedytosafety(n)	(hazardtosafety(remedytohazard(n)))
#define safetytohazard(n)	((n) + (TC_OUTOFGAS - TC_EXTRATANK))


/*
 * Data about a player
 */
typedef	struct	PLAYER	PLAYER;

struct PLAYER	{
	void	(*p_move) PROTO((PLAYER *));	/* play normal move proc */
	BOOL	(*p_extend) PROTO((PLAYER *));	/* ask for extended hand proc */
	PLAYER	*p_enemy;		/* opposing player */
	int	p_index;		/* index of player */
	BOOL	p_tripdone;		/* TRUE if player completed trip */
	BOOL	p_moved;		/* TRUE if player has finished his turn */
	BOOL	p_stuck;		/* TRUE if player is totally stuck */
	BOOL	p_stopped;		/* TRUE if player is stopped */
	BOOL	p_limited;		/* TRUE if player has speed limit */
	int	p_stopturns;		/* number of turns we have been stopped */
	int	p_limitturns;		/* number of turns we have been limited */
	int	p_score;		/* score so far this hand */
	int	p_gamescore;		/* score so far this game */
	int	p_games;		/* games won */
	int	p_mileage;		/* mileage played so far this hand */
	CARD	p_battlepile;		/* card on top of battle pile */
	CARD	p_oldbattle;		/* previous card on battle pile */
	CARD	p_speedpile;		/* card on top of speed limit pile */
	CARD	p_oldspeed;		/* previous card on speed pile */
	CARD	p_coupcard;		/* hazard that can be coup fourred */
	int	p_cardcount;		/* number of cards in hand */
	int	p_pickindex;		/* index of card which was just picked */
	int	p_shutouts;		/* total number of shutout hands */
	int	p_wonhands;		/* total hands won */
	int	p_besthand;		/* highest hand score to date */
	int	p_bestgame;		/* highest game score to date */
	int	p_worstgame;		/* lowest game score to date */
	long	p_allgamescore;		/* total scores for all games */
	int	p_tengamescores[10];	/* scores of last ten games */
	int	p_spare[83];		/* spare words for expansion */
	CARD	p_cards[HANDSIZE+1];	/* cards in player's hand */
	int	p_milecards[MILETYPES];	/* count of mileage cards played */
	int	p_safecards[SAFETYPES];	/* count of safety cards played */
	int	p_coupcards[SAFETYPES];	/* count of coup fourres played */
};


/*
 * Data written or read when a game is saved or restored
 */
typedef	struct	{
	long	sd_magic;		/* magic number for validity */
	PLAYER	sd_players[2];		/* player data */
	CARD	sd_carddeck[CARDCOUNT];	/* cards remaining in deck */
	int	sd_typesleft[CARDTYPES]; /* number of each type of card left */
	int	sd_games;		/* total games played */
	int	sd_hands;		/* total hands played */
	int	sd_curhelp;		/* current help screen */
	int	sd_curindex;		/* current player index */
	STATUS	sd_curstatus;		/* current game status */
	int	sd_mileagegoal;		/* mileage goal */
	int	sd_cardsleft;		/* total number of cards left */
	CARD	sd_discardpile;		/* card on top of discard pile */
	int	sd_spare[97];		/* room for expansion */
	long	sd_checksum;		/* checksum of data (MUST BE LAST) */
} SAVEDATA;

#define SAVESIZE (sizeof(SAVEDATA))	/* size of save stucture */
#define SUMSIZE (SAVESIZE - sizeof(long))	/* amount of data to checksum */


/*
 * Defines to make references to information in savedata easy.
 */
#define discardpile	savedata.sd_discardpile
#define mileagegoal	savedata.sd_mileagegoal
#define cardsleft	savedata.sd_cardsleft
#define players		savedata.sd_players
#define typesleft	savedata.sd_typesleft
#define carddeck	savedata.sd_carddeck
#define curindex	savedata.sd_curindex
#define curstatus	savedata.sd_curstatus
#define curhelp		savedata.sd_curhelp


/*
 * Global data.
 */
extern	char	*cardnames[CARDTYPES];	/* card names */
extern	char	*carddebug[CARDTYPES];	/* short names of cards for debugging */
extern	int	cardcounts[CARDTYPES];	/* number of each card in game */
extern	int	cardranks[CARDTYPES];	/* sorting order of cards in hand */
extern	int	mileage[MILETYPES];	/* mileages for mileage cards */
extern	SAVEDATA	savedata;	/* save file format */
extern	PLAYER	*player;		/* data about real player */
extern	PLAYER	*machine;		/* data about machine */
extern	char	*savefile;		/* filename for save file */
extern	BOOL	debug;			/* debugging flag */


/*
 * Global routines
 */
extern	char	*canplaymsg PROTO((PLAYER *, CARD));
extern	void	playcard PROTO((PLAYER *, CARD));
extern	void	checkstop PROTO((PLAYER *));
extern	CARD	fixcard PROTO((PLAYER *));
extern	void	discard PROTO((PLAYER *, CARD));
extern	STATUS	gamestatus PROTO((void));
extern	void	score PROTO((PLAYER *));
extern	char	*pickcards PROTO((PLAYER *, int));
extern	void	sorthand PROTO((PLAYER *));
extern	void	play_move PROTO((PLAYER *));
extern	BOOL	play_extend PROTO((PLAYER *));
extern	BOOL	play_again PROTO((int));
extern	void	mach_move PROTO((PLAYER *));
extern	BOOL	mach_extend PROTO((PLAYER *));
extern	void	showgame PROTO((void));
extern	void	nexthelp PROTO((void));
extern	char	*savegame PROTO((char *));
extern	char	*restoregame PROTO((char *));
extern	void	message PROTO((char *));
extern	BOOL	yesno PROTO((char *, char *));
extern	char	*readline PROTO((char *, char *, int));


/*
 * Macro to check to see if playing a card is legal
 */
#define canplay(pp, card)	(canplaymsg(pp, card) == NULL)

/* END CODE */
